/*
 * BMCpp
 * Copyright (C) 2016, 2017 Wilhelm M <guessmail>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <cstdint>

#include "avr8defs.h"

#pragma pack(push)
#pragma pack(1)

namespace AVR {
    struct ATMega328P final {
        // most definitions excluded ;-)
        
        template<typename T>
        static constexpr bool is_atomic() {return false;}
        
        ATMega328P() = delete;
        struct Usart {
            static constexpr const uint8_t count = 1;
            enum class UCSRA : uint8_t {
                rxc = (1 << RXC0),
                txc = (1 << TXC0),
                udre = (1 << UDRE0),
                fe = (1 << FE0),
                dor = (1 << DOR0),
                upe = (1 << UPE0),
                u2x = (1 << U2X0),
                mpcm = (1 << MPCM0)
            };
            ControlRegister<Usart, UCSRA> ucsra;
            enum class UCSRB : uint8_t {
                rxcie = (1 << RXCIE0),
                txcie = (1 << TXCIE0),
                udrie = (1 << UDRIE0),
                rxen = (1 << RXEN0),
                txen = (1 << TXEN0),
                ucsz2 = (1 << UCSZ02),
                rxb8 = (1 << RXB80),
                txb8 = (1 << TXB80)
            };
            ControlRegister<Usart, UCSRB> ucsrb;
            enum class UCSRC : uint8_t {
                umsel0 = (1 << UMSEL00),
                umsel1 = (1 << UMSEL01),
                upm1 = (1 << UPM01),
                upm0 = (1 << UPM00),
                usbs = (1 << USBS0),
                ucsz0 = (1 << UCSZ00),
                ucsz1 = (1 << UCSZ01),
                ucpol = (1 << UCPOL0)
            };
            ControlRegister<Usart, UCSRC> ucsrc;
            volatile uint8_t reserved1;
            DataRegister<Usart, ReadWrite, uint16_t> ubbr;
            DataRegister<Usart, ReadWrite, std::byte> udr;
            volatile uint8_t reserved2;
            template<int N> struct Address;
        };
        struct PortRegister {
            DataRegister<PortRegister, ReadWrite, std::byte> in; // ReadWrite due to toggle RMW
            DataRegister<PortRegister, ReadWrite, std::byte> ddr;
            DataRegister<PortRegister, ReadWrite, std::byte> out;
            template<typename P> struct Address;
        };
        struct Interrupt {
            enum class PCFlags : uint8_t {
                if2 = (1 << PCIF2),
                if1 = (1 << PCIF1),
                if0 = (1 << PCIF0)
            };
            FlagRegister<Interrupt, PCFlags> pcifr;
            volatile uint8_t eifr;
            volatile uint8_t eimsk;
            volatile uint8_t padding[0x68 - 0x3b - 1 - 2];
            enum class PCMask : uint8_t {
                ie2 = (1 << PCIE2),
                ie1 = (1 << PCIE1),
                ie0 = (1 << PCIE0)
            };
            ControlRegister<Interrupt, PCMask> pcicr;
            volatile uint8_t eicra;
            static constexpr uint8_t address = 0x3b;
        };
        struct GPIOR final {
            static constexpr const uint8_t count = 3;
            template<uint8_t N> struct Address; 
        };
        struct Status final {
            enum class Bits : uint8_t {
                globalIntEnable = (1 << 7), 
                bitCopy = (1 << 6) 
            };
            ControlRegister<Status, Bits> value;
            static constexpr uint8_t address = 0x5f;
        };
    };
    template<>
    constexpr bool ATMega328P::is_atomic<uint8_t>() {return true;}
}

namespace std {
    template<>
    struct enable_bitmask_operators<AVR::ATMega328P::Status::Bits> {
        static constexpr bool enable = true;
    };
    template<>
    struct enable_bitmask_operators<AVR::ATMega328P::Usart::UCSRA> {
        static constexpr bool enable = true;
    };
    template<>
    struct enable_bitmask_operators<AVR::ATMega328P::Usart::UCSRB> {
        static constexpr bool enable = true;
    };
    template<>
    struct enable_bitmask_operators<AVR::ATMega328P::Usart::UCSRC> {
        static constexpr bool enable = true;
    };
}

namespace AVR {
    
    template<>
    struct ATMega328P::GPIOR::Address<0> {
        inline static constexpr uintptr_t value = 0x3e;
    };
    template<>
    struct ATMega328P::GPIOR::Address<1> {
        inline static constexpr uintptr_t value = 0x4a;
    };
    template<>
    struct ATMega328P::GPIOR::Address<2> {
        inline static constexpr uintptr_t value = 0x4b;
    };
    
    template<>
    struct ATMega328P::PortRegister::Address<B> {
        static constexpr uint8_t value = 0x23;
    };
    template<>
    struct ATMega328P::PortRegister::Address<C> {
        static constexpr uint8_t value = 0x26;
    };
    template<>
    struct ATMega328P::PortRegister::Address<D> {
        static constexpr uint8_t value = 0x29;
    };
    
    template<>
    struct ATMega328P::Usart::Address<0> {
        static constexpr uint8_t value = 0xc0;
    };
    
    
}
#pragma pack(pop)
