/*
 * BMCpp
 * Copyright (C) 2016, 2017 Wilhelm M <guessmail>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <cstdint>
#include <type_traits>

namespace std {
    
    template<typename C>
    concept bool Clock() {
        return requires(C clock) {
            clock.dateTime();
        };
    }
    
    template<typename T>
    concept bool Integral() {
        return std::is_integral<T>::value;    
    }
    
    template<typename T>
    concept bool Unsigned() {
        return std::is_unsigned<T>::value;    
    }
    
    template<typename T>
    concept bool Signed() {
        return std::is_signed<T>::value;    
    }
    
    template<typename R>
    concept bool Range() { 
        return requires (R r) { 
            typename R::value_type;
            r.begin();
            r.end();
        };
    }
    template<typename T>
    concept bool Fundamental()
    {
        return std::is_fundamental<T>::value;
    }
    
    template<typename T>
    concept bool NonFundamental()
    {
        return !std::is_fundamental<T>::value;
    }
    
}
